/*
 * Copyright (c) 2005 Silicon Graphics, Inc.
 * All rights reserved.
 *
 *   Michael A. Raymond <mraymond@sgi.com>
 */

/*
 *  This file is part of the ULI sample applications.
 *
 *  uli_sample is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  uli_sample is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with uli_sample; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>
#include <uli.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>

/*
 * This is a simple demonstration application which uses the User
 * Level Interrupt (ULI) feature.
 */

#define ULI_CPU 0	/* CPU to use */
#define THREAD_SEMA 0	/* Sema thread waits on */
#define NUM_SEMAS 1		/* Number of semaphores to use */

static int uli_data = 0;
static void * intr;
pthread_attr_t attrs;
pthread_t pt;

/*
 * This is the handler function for the ULI.  It modifies a global
 * variable then wakes up a waiter.  It demonstrates a simple
 * handler plus the use of ULI_wakeup.
 */
static void
uli_handler(void * arg_data)
{
	int * data = (int*) arg_data;

	/* Increment some global counter */
	(*data)++;

	/* Wake up the waiting thread */
	ULI_wakeup(intr, THREAD_SEMA);

	/* We're done, complete */
	return;
}

/*
 * This simple thread waits for the ULI to fire and then exits.  It
 * demonstrates the use of the ULI_sleep.
 */
static void *
thread_func(void * arg)
{
	printf("Thread blocking on ULI\n");

	if (ULI_sleep(arg, THREAD_SEMA) < 0) {
		perror("ULI_sleep");
		exit(1);
	}

	printf("Thread woken by ULI and exiting\n");

	return NULL;
}

int
main(int argc, char * argv[])
{
	int line;
	int before;

	/* Check for the needed input argument */
	if (argc <= 1) {
		fprintf(stderr, "Need to specificy an IRQ line to use\n");
		return 1;
	}

	/* Get the IRQ line to use */
	line = atoi(argv[1]);

	/* Lock down memory */
	if (mlockall(MCL_CURRENT | MCL_FUTURE) < 0) {
		perror("mlockall");
		return 1;
	}

	/*
	 * Our ULI handler calls ULI_wakeup().  glibc for some reason
	 * uses floating point registers when resolving library symbols.
	 * We therefor pre-resolve ULI_wakeup before registering the
	 * handler.  There must be a better way...
	 */
	ULI_wakeup(NULL, -1);

	/*
	 * Create the ULI.  We create NUM_SEMAS semaphores and have the
	 * library create a default size stack.
	 */
	if ((intr = ULI_register_irq(ULI_CPU, uli_handler, &uli_data, NUM_SEMAS,
					 NULL, 0, line)) == NULL) {
		perror("ULI_register_irq");
		return -1;
	}

	/* Create a child thread */
	pthread_attr_init(&attrs);
	if (pthread_create(&pt, &attrs, thread_func, intr)) {
		perror("pthread_create");
		return 1;
	}

	/* Wait for it to get woken by the ULI and exit */
	pthread_join(pt, NULL);

	/* Let the user know we're checking blocking */
	printf("Child completed\n");
	printf("Checking interrupt blocking\n");

	/* Block the ULI handler from running */
	ULI_block_intr(intr);

	/* Check to make sure that the ULI actually is blocked */
	before = uli_data;
	sleep(5);
	printf("ULIs handled before blocking: %d\n",
		before);
	printf("ULIs handled while blocking: %d\n",
		uli_data - before);

	/* Unblock the ULI handler */
	ULI_unblock_intr(intr);

	/* Destroy the ULI */
	if (ULI_destroy(intr) < 0) {
		perror("ULI_destroy");
		return -1;
	}

	/* Let's see how many times the ULI got handled */
	printf("ULI handler ran %d times\n",
		uli_data);

	return 0;
}
